"""
_pg_root.py
04. February 2023

the root of the window

Author:
Nilusink
"""
from .widgets import GeometryManager
from .theme import ThemeManager
import pygame as pg
import typing as tp

from .types import *


DEFAULT_TITLE: str = "Window"
DEFAULT_ICON: str = "icon.png"


class PgRoot(GeometryManager):
    _running: bool = True
    _theme: ThemeManager = ...
    __background: pg.Surface = ...

    def __init__(
            self,
            title: str = ...,
            icon_path: str = ...,
            size: tuple[int, int] = ...,
            bg_color: Color = ...,
    ):
        super().__init__()
        self._theme = ThemeManager()

        # args
        self._bg = self._theme.root.bg.hex if bg_color is ... else bg_color

        # pg init
        pg.init()
        pg.font.init()

        self.__background = pg.display.set_mode()

        # set icon and caption
        pg.display.set_caption(DEFAULT_TITLE if title is ... else title)
        img = pg.image.load(DEFAULT_ICON if icon_path is ... else icon_path, "icon")
        pg.display.set_icon(img)

    # config
    @property
    def title(self) -> str:
        return pg.display.get_caption()[0]

    @title.setter
    def title(self, value: str) -> None:
        pg.display.set_caption(value)

    @property
    def theme(self) -> ThemeManager:
        return self._theme

    # pygame stuff
    def _event_handler(self) -> None:
        """
        handle the events raised by pygame
        """
        for event in pg.event.get():
            match event.type:
                case pg.QUIT:
                    self._running = False

    def update_screen(self) -> None:
        """
        update the screen
        """
        self.__background.fill(self._bg)

        self.calculate_geometry()
        for child, params in self._child_params:
            child.draw(self.__background)

        pg.display.flip()

    def update(self) -> None:
        """
        update events
        """
        self._event_handler()

    def mainloop(self):
        """
        run the windows main loop
        """
        while self._running:
            self.update()
            self.update_screen()

    def calculate_geometry(self):
        """
        calculate how each individual child should be placed
        """
        match self._layout:
            case 0:  # Absolute
                # since the positioning is absolute, the children should not influence the parents size
                for child, params in self._child_params:
                    child.set_position(params.x, params.y)

                return

            case 1:
                ...

            case 2:
                ...

            case _:
                raise ValueError(f"Invalid geometry type: {self._layout.__class__.__name__}")

    def calculate_size(self) -> tuple[int, int]:
        """
        calculate how big the container should be
        """
        # make sure the geometry is up-to-date
        self.calculate_geometry()

        return pg.display.get_window_size()



"""
_frame.py
04. February 2023

Frame - the base widget

Author:
Nilusink
"""
from ._geo_manager import GeometryManager
from ..types import Color, BetterDict
import typing as tp
import pygame as pg


class DisplayConfig(tp.TypedDict):
    bg: Color
    ulr: int  # border radii
    urr: int
    llr: int
    lrr: int


class Frame(GeometryManager):
    """
    The base widget
    """
    __parent: tp.Union["Frame", tp.Any]
    _display_config: BetterDict[DisplayConfig] = ...
    _x: int = -1
    _y: int = -1

    def __init__(
            self,
            parent: tp.Union["Frame", tp.Any],
            width: int = ...,
            height: int = ...,
            bg_color: Color = ...,
            border_radius: int = ...,
    ) -> None:
        # mutable defaults
        self._display_config = BetterDict({
            "bg": Color(),
            "ulr": 0,
            "urr": 0,
            "llr": 0,
            "lrr": 0
        })

        super().__init__()

        # arguments
        self.__parent = parent

        if width is not ...:
            self._width = width

        if height is not ...:
            self._height = height

        self._display_config["bg"] = self.__parent.theme.frame.bg if bg_color is ... else bg_color

        if border_radius:
            self._display_config["ulr"] = self._display_config["urr"] = border_radius
            self._display_config["llr"] = self._display_config["lrr"] = border_radius

    def get_size(self) -> tuple[int, int]:
        """
        get the frames size (including children)
        """
        calculated_size = self.calculate_size()

        # width
        width = self._width
        if width == -1:
            width = calculated_size[0]
            width = 0 if width == -1 else width

        # height
        height = self._height
        if height == -1:
            height = calculated_size[1]
            height = 0 if height == -1 else height

        return width.__floor__(), height.__floor__()

    def draw(self, surface: pg.Surface) -> None:
        """
        draw the frame
        """
        width, height = self.get_size()
        _surface = pg.Surface((width, height), pg.SRCALPHA)

        # draw the frame
        pg.draw.rect(_surface, self._display_config.bg.rgba, pg.Rect((0, 0, self._width, self._height)))

        # draw children
        for child, params in self._child_params:
            child.draw(_surface)

        # draw
        surface.blit(_surface, (self._x, self._y))

    def place(
            self,
            x: int,
            y: int,
    ) -> None:
        """
        place the frame in a parent container
        """
        self.__parent.add_child(self, x=x, y=y)

    def set_position(self, x: int, y: int) -> None:
        """
        set the child's position (used by parents)
        """
        self._x = x
        self._y = y

    def set_size(self, width: float, height: float) -> None:
        """
        set the child's size (used by parens)
        """
        self._width = width
        self._height = height



"""
_better_dict.py
04. February 2023

A dict type similar to the javascript JSON object

Author:
Nilusink
"""


class BetterDict(dict):
    def __getattr__(self, item):
        result = self[item]
        if issubclass(type(result), dict):
            return BetterDict(result)

        return result

    def __setattr__(self, key, value):
        self[key] = value



"""
_color.py
04. February 2023

a color class

Author:
Nilusink
"""
import typing as tp


class Color:
    _rgb: tuple[float, float, float] = ...
    _alpha: float = 255
    _hex: str = ...

    def __init__(self) -> None:
        self.rgb = 0, 0, 0, 0

    @classmethod
    def from_rgb(cls, red: float, green: float, blue: float, alpha: float = ...) -> "Color":
        new = cls()
        new.rgb = [red, green, blue] + ([] if alpha is ... else [alpha])

        return new

    @classmethod
    def from_hex(cls, hex_value: str, alpha: float = ...) -> "Color":
        new = cls()
        new.hex = hex_value
        new.alpha = alpha

        return new

    @property
    def rgb(self) -> tuple[float, float, float]:
        return self._rgb

    @property
    def rgba(self) -> tuple[float, float, float, float]:
        return tuple(self._rgb) + (self._alpha,)

    @rgb.setter
    def rgb(self, value: tuple[float, float, float] | tuple[float, float, float, float]) -> None:
        if len(value) == 4:
            *self._rgb, self._alpha = value

        elif len(value) == 3:
            self._rgb = value

        else:
            raise ValueError(f"The rgb tuple must have a length of either 3 or 4!  (not {len(value)})")

        self._recalculate_colors("r")

    @property
    def hex(self) -> str:
        return self._hex

    @hex.setter
    def hex(self, value: str) -> None:
        if len(value) == 3:  # ex: fff
            self._hex = "#" + 2*value[0] + 2*value[1] + 2*value[2]

        elif len(value) == 4:  # ex: #fff
            self._hex = "#" + 2*value[1] + 2*value[2] + 2*value[3]

        elif len(value) == 6:  # ex: ffffff
            self._hex = "#" + value

        elif len(value) == 7:  # ex: #ffffff
            self._hex = value

        else:
            raise ValueError(f"The hex string must be either 3, 4 or 6 or 7 characters long! (not {len(value)})")

        self._recalculate_colors("h")

    @property
    def alpha(self) -> float:
        return self.alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        self._alpha = value

    def _recalculate_colors(self, calc_from: tp.Literal["r", "rgb", "h", "hex"]):
        if calc_from in ("rgb", "r"):
            self._hex = f"#" \
                        f"{hex(self.rgb[0].__floor__())[2:]}" \
                        f"{hex(self.rgb[1].__floor__())[2:]}" \
                        f"{hex(self.rgb[2].__floor__())[2:]}".upper()

        elif calc_from in ("hex", "h"):
            r = self._hex[1:3]
            g = self._hex[3:5]
            b = self._hex[5:7]

            self._rgb = (
                int(r, 16),
                int(g, 16),
                int(b, 16)
            )

            self._hex.upper()

    # magic
    def __hex__(self) -> str:
        return self.hex

    def __repr__(self) -> str:
        return f"<{self.hex} - {self.rgb}>"

    def __str__(self) -> str:
        return self.__repr__()



"""
_manager.py
04. February 2023

The built-in theme manager

Author:
Nilusink
"""
from ..types import *
import typing as tp
import json
import os


DEFAULT_THEME: str = os.path.dirname(__file__) + "/themes/default.json"


class ThemeManager:
    """
    the build-in theme manager
    """
    _config: dict[str, str | dict] = ...
    instance = ...

    def __new__(cls, *args, **kwargs):
        """
        only on instance of a theme manager should ever exist
        """
        if cls.instance is ...:
            cls.instance = super().__new__(cls, *args, **kwargs)

        return cls.instance

    def __init__(self, theme_path: str = ...) -> None:
        # load the theme
        self._config = json.load(open(DEFAULT_THEME if theme_path is ... else theme_path, "r"))

        # convert each color to a color class instance
        for key in self._config.copy():
            for ckey, color in self._config[key].items():
                if isinstance(color, str):
                    # differentiate between hex and rgb values
                    if color.startswith("#"):
                        self._config[key][ckey] = Color.from_hex(color, 255)

                    elif color.startswith("rgb"):
                        self._config[key][ckey] = Color.from_rgb(*[int(val) for val in color.lstrip("rgb").split(",")])

                    else:
                        raise ValueError(f"Invalid color value in theme file: \"{color}\"")

                # rgb values written as tuple
                elif isinstance(color, tp.Iterable) and isinstance(color[0], float) or isinstance(color[0], int):
                    self._config[key][ckey] = Color.from_rgb(*color)

                else:
                    raise ValueError(f"Invalid value in theme file: {color}")

    def __getattr__(self, item: str) -> str | BetterDict:
        """
        for better accessibility
        """
        val = self._config[item]

        if isinstance(val, dict):
            return BetterDict(val)

        elif isinstance(val, str):
            return val

        else:
            raise ValueError(f"Invalid type for key \"{item}\": {type(val)}")

    def __getitem__(self, item: str) -> str | BetterDict:
        return self.__getattr__(item)




"""
_geo_manager.py
04. February 2023

How children are placed

Author:
Nilusink
"""
from ..types import Absolute, Pack, Grid, BetterDict
from ._supports_children import SupportsChildren
import typing as tp


class GeometryManager(SupportsChildren):
    """
    Manages how children are placed inside a parent container
    """
    _layout: int = Absolute
    _width: float = -1   # -1 means not configured -> takes minimum size required by its children
    _height: float = -1  # or the size it gets by the parents geometry manager
    _child_params: list[tuple[tp.Any, BetterDict]] = ...

    def __init__(self, layout: int = ...):
        super().__init__()

        self._layout = Absolute if layout is ... else layout
        self._child_params = []

    def set_layout(self, layout: int) -> None:
        """
        set the container's layout type
        """
        if layout not in (Absolute, Pack, Grid):
            raise ValueError("Invalid container layout: ", layout)

        self._layout = layout

    def add_child(self, child: tp.Any, **params) -> None:
        """
        add a child to the collection
        """
        if child not in self._children:
            super().add_child(child)
            self._child_params.append((child, BetterDict(params)))

    def calculate_geometry(self):
        """
        calculate how each individual child should be placed
        """
        match self._layout:
            case 0:  # Absolute
                # since the positioning is absolute, the children should not influence the parents size
                for child, params in self._child_params:
                    child.set_position(params.x, params.y)

                return

            case 1:
                ...

            case 2:
                ...

            case _:
                raise ValueError(f"Invalid geometry type: {self._layout.__class__.__name__}")

    def calculate_size(self) -> tuple[int, int]:
        """
        calculate how big the container should be
        """
        # make sure the geometry is up-to-date
        self.calculate_geometry()

        match self._layout:
            case 0:  # Absolute
                # since the positioning is absolute, the children should not influence the parents size
                return self._width.__floor__(), self._height.__floor__()

            case 1:
                ...

            case 2:
                ...

            case _:
                raise ValueError(f"Invalid geometry type: {self._layout.__class__.__name__}")
